import os

from dodo_commands.dependencies.get import funcy
from dodo_commands.framework.funcy import remove_if

concat, distinct = funcy.concat, funcy.distinct


def get_ordered_layer_paths(ctr):
    root_layer_path = ctr.layers.config_io.glob([ctr.layers.root_layer_path
                                                 ])[0]

    x = concat(ctr.layers.selected_layer_by_path.keys(),
               ctr.layers.layer_by_target_path.keys())
    x = distinct(x)
    x = sorted(x, key=os.path.basename)
    x = ctr.layers.config_io.glob(x)

    x = remove_if(lambda x: x == root_layer_path)(x)
    x = concat([root_layer_path], x)

    return x



import sys

from dodo_commands.framework.config_expander import Key


class ConfigArg:
    def __init__(self, config_key, *args, **kwargs):
        self.config_key = config_key
        self.args = args
        self.kwargs = kwargs

    @property
    def arg_name(self):
        return self.args[0].strip('-').replace('-', '_')


def add_config_args(parser, config, config_args):
    show_help = '--help' in sys.argv

    for config_arg in (config_args or []):
        key = Key(config, config_arg.config_key)
        key_exists = key.exists()

        if show_help or not key_exists:
            kwargs = dict(config_arg.kwargs)
            help_text = kwargs.get('help') or ''
            sep = '. ' if help_text else ''
            if key_exists:
                value = str(key.get())
                formatted_value = value[:50] + (value[50:] and '...')
                extra_help = ('Read from config: %s = %s.' %
                              (config_arg.config_key, formatted_value))
            else:
                extra_help = ('Configuration key is %s' %
                              config_arg.config_key)

            kwargs['help'] = help_text + sep + extra_help
            parser.add_argument(*config_arg.args, **kwargs)



import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum, six

docker = plumbum.cmd.docker
raw_input = six.moves.input


def _args():
    parser = ArgumentParser()
    parser.add_argument('--by-container-name', dest='container_name')
    args = Dodo.parse_args(parser)
    return args


def _containers():
    result = []
    for line in docker("ps", "--format",
                       "{{.ID}} {{.Names}} {{.Image}}").split('\n'):
        if line:
            cid, name, image = line.split()
            result.append(dict(name=name, cid=cid, image=image))
    return result


if Dodo.is_main(__name__):
    args = _args()
    containers = _containers()

    if args.container_name:
        filtered_containers = [
            x for x in containers if x.get('name') == args.container_name
        ]
        if not filtered_containers:
            raise CommandError("Container not found: %s" % args.container_name)
        assert len(filtered_containers) == 1
        container = filtered_containers[0]
    else:
        print("0 - exit")
        for idx, container in enumerate(containers):
            print("%d - %s" % (idx + 1, container['name']))

        print("\nSelect a container: ")
        choice = int(raw_input()) - 1

        if choice == -1:
            sys.exit(0)

        container = containers[choice]

    Dodo.run([
        'docker',
        'commit',
        container['cid'],
        container['image'],
    ], )



import os
import sys

from dodo_commands.framework.command_map import CommandMapItem


class ShellCommandMapItem(CommandMapItem):
    def __init__(self, group, filename):
        super(ShellCommandMapItem, self).__init__(group, filename, 'sh')


class ShellCommandHandler:
    def add_commands_to_map(self, command_dirs, file_map, command_map):
        for command_dir, files in file_map.items():
            for file in files:
                command_name, ext = os.path.splitext(os.path.basename(file))
                prefix = 'dodo.'
                if ext == '.sh' and command_name.startswith(prefix):
                    command_name = command_name[len(prefix):]
                    command_map[command_name] = ShellCommandMapItem(
                        group=os.path.basename(command_dir), filename=file)

    def execute(self, command_map_item, command_name):
        from dodo_commands.framework.singleton import Dodo
        Dodo.run([command_map_item.filename] + sys.argv[2:])



def transform_prefixes_in_aliases(config, layer_props_by_layer_name):
    aliases = config["ROOT"].get("aliases", {})
    for name, value in list(aliases.items()):
        parts = value.split()
        for idx, part in list(enumerate(parts)):
            if part.startswith("--"):
                continue

            prefixes = part.split(".")
            parts[idx] = prefixes.pop()

            for prefix in prefixes:
                layer = (layer_props_by_layer_name[prefix].target_path
                         if prefix in layer_props_by_layer_name else prefix)
                parts.append("--layer=%s" % layer)

            aliases[name] = " ".join(parts)
            break



import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum, six
from dodo_commands.framework.choice_picker import ChoicePicker
from dodo_commands.framework.util import query_yes_no

docker = plumbum.cmd.docker
raw_input = six.moves.input


def _container_type_names():
    return list(Dodo.get_config('/DOCKER/container_types', {}).keys())


def _args():
    parser = ArgumentParser()
    parser.add_argument('container_type',
                        choices=_container_type_names(),
                        nargs='?')
    parser.add_argument('name', nargs='?')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--interactive', '-i', action='store_true')
    group.add_argument('--replace',
                       action='store_true',
                       help='Replace existing container')

    args = Dodo.parse_args(parser)
    args.container_types = Dodo.get_config("/DOCKER/container_types")
    args.env_name = Dodo.get_config('/ROOT/env_name')

    return args


def _exists(name):
    return docker("ps", "-a", "--quiet", "--filter", "name=" + name)


def _create_container(container_types, container_type_name, name, replace):
    dirs = container_types.get(container_type_name, {})['dirs']
    image = container_types.get(container_type_name, {})['image']

    exists = _exists(name)

    if exists and replace:
        Dodo.run(['docker', 'rm', name])

    if not exists or replace:
        cmd_args = [
            "docker",
            "create",
            "--name",
            name,
        ]

        for path in dirs:
            cmd_args.extend(["-v", path])

        cmd_args.append(image)
        Dodo.run(cmd_args)


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if args.interactive:

        class Picker(ChoicePicker):
            def print_choices(self, choices):
                for idx, container_type in enumerate(choices):
                    print('%d - %s' % (idx + 1, container_type))

            def question(self):
                return "Select a container type: "

        picker = Picker(_container_type_names())
        picker.pick()

        for container_type_name in picker.get_choices():
            default_container_name = 'dc_%s_%s' % (args.env_name,
                                                   container_type_name)
            name = raw_input(
                "Enter a name for container type %s [%s]: " %
                (container_type_name,
                 default_container_name)) or default_container_name

            if _exists(name):
                replace = query_yes_no(
                    'A container with this name exists. ' +
                    'To replace choose yes. To quit choose no. Replace it?')
                if not replace:
                    sys.exit(0)

            _create_container(args.container_types,
                              container_type_name,
                              name,
                              replace=True)
    else:
        if not args.container_type:
            raise CommandError('Argument container_type is required')
        if not args.name:
            raise CommandError('Argument name is required')
        _create_container(args.container_types, args.container_type, args.name,
                          args.replace)



from dodo_commands.framework.command_map import get_command_map
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.container.facets import (Commands, Config, Layers,
                                                      i_, map_datas, o_)
from dodo_commands.framework.funcy import drill
from dodo_commands.framework.inferred_commands import get_inferred_command_map


# COMMANDS
def action_get_inferred_command_map(ctr):
    def transform(
        global_aliases,
        layer_props_by_layer_name,
        layer_by_target_path,
    ):
        layer_name_by_inferred_command = get_inferred_command_map(
            global_aliases,
            layer_props_by_layer_name,
            layer_by_target_path,
        )
        return (layer_name_by_inferred_command, )

    return map_datas(i_(Commands, 'global_aliases'),
                     i_(Layers, 'layer_props_by_layer_name'),
                     i_(Layers, 'layer_by_target_path'),
                     o_(Commands, 'layer_name_by_inferred_command'),
                     transform=transform)(ctr)


# COMMANDS
def action_get_aliases_from_config(ctr):
    def transform(config):
        return (drill(config, 'ROOT', 'aliases', default={}), )

    return map_datas(i_(Config, 'config'),
                     o_(Commands, 'aliases_from_config'),
                     transform=transform)(ctr)


# COMMANDS
def action_get_command_map(ctr):
    def transform(config):
        command_dirs = get_command_dirs_from_config(config)
        command_map = get_command_map(command_dirs)
        return (command_dirs, command_map)

    return map_datas(i_(Config, 'config'),
                     o_(Commands, 'command_dirs'),
                     o_(Commands, 'command_map'),
                     transform=transform)(ctr)



import os
from argparse import ArgumentParser

from dodo_commands import Dodo


def _args():
    parser = ArgumentParser(description=(
        "Writes (or removes) a small script that activates the latest " +
        "Dodo Commands project"))
    parser.add_argument('status', choices=['on', 'off'])
    return Dodo.parse_args(parser)


if Dodo.is_main(__name__, safe=False):
    args = _args()

    for shell, activate_cmd in (("bash", "$(dodo env --latest)"),
                                ("fish", "source (dodo env --latest)")):

        confd_dir = os.path.expanduser("~/.config/%s/conf.d" % shell)
        if not os.path.exists(confd_dir):
            Dodo.run(["mkdir", "-p", confd_dir])
        script = os.path.join(confd_dir, "dodo_autostart")
        if args.status == "on" and not os.path.exists(script):
            with open(script, "w") as f:
                f.write("# NOTE: automatically generated file, don't edit.\n")
                f.write("%s && dodo check-version --dodo --config\n" %
                        activate_cmd)
        if args.status == "off" and os.path.exists(script):
            os.unlink(script)



import argparse
import os
import sys

from dodo_commands.dependencies.get import funcy
from dodo_commands.framework.funcy import keep_truthy, str_split_at

cut_suffix, distinct = funcy.cut_suffix, funcy.distinct


class CommandLine:
    def __init__(self):
        self.input_args = None
        self.cleaned = None
        self.command_name = None
        self.is_help = False
        self.is_trace = False
        self.inferred_layer_paths = []
        self.expanded_layer_paths = []
        self.given_layer_paths = []
        self.more_given_layer_paths = []
        self.is_running_directly_from_script = False

    @property
    def layer_paths(self):
        return distinct(self.inferred_layer_paths + self.expanded_layer_paths +
                        self.given_layer_paths + self.more_given_layer_paths)

    @property
    def _split_input1(self):
        input1 = self.input_args[1] if len(self.input_args) > 1 else ""

        # When running directly from a script, then we know the command was not
        # prefixed with layer names.
        if self.is_running_directly_from_script:
            return '', cut_suffix(input1, '.py')
        return str_split_at(input1, input1.rfind('.') + 1)

    @property
    def raw_command_name(self):
        return self._split_input1[1]

    @property
    def raw_command_prefix(self):
        return self._split_input1[0]

    @property
    def layer_names(self):
        return distinct(keep_truthy()(self.raw_command_prefix.split('.')))

    def get_trace(self):
        args = [x for x in self.input_args if x != '--trace']
        args[1] = self.command_name
        args.extend(["--layer=%s" % x for x in self.layer_paths])
        return args

    @staticmethod
    def get(ctr):
        return ctr.command_line


def init_command_line(self):
    self.is_running_directly_from_script = '__DODO__' not in os.environ
    parser = argparse.ArgumentParser()
    args = (
        # If a dodo script was called directly via the python interpreter,
        # then add sys.executable to make all calls look similar
        [sys.executable, *sys.argv]
        if self.is_running_directly_from_script else
        #
        sys.argv)

    if '--help' in args:
        if not ('--' in args and args.index('--') < args.index('--help')):
            self.is_help = True
            args = [x for x in args if x != '--help']

    parser.add_argument('-L', '--layer', action='append')
    parser.add_argument('--trace', action='store_true')

    known_args, args = parser.parse_known_args(args)

    self.given_layer_paths = known_args.layer or []
    self.is_trace = known_args.trace
    self.input_args = args

    return self



from dodo_commands.framework.singleton import Dodo


# Resp: add the current command_name
# to the list of commands decorated by decorator_name.
class DecoratorScope:
    def __init__(self, decorator_name):
        self.decorators = Dodo.get_config('/ROOT').setdefault(
            'decorators', {}).setdefault(decorator_name, [])

    def __enter__(self):  # noqa
        self.decorators.append(Dodo.command_name)

    def __exit__(self, type, value, traceback):  # noqa
        self.decorators.remove(Dodo.command_name)



import glob
import os


class CommandMapItem(object):
    def __init__(self, group, filename, extension):
        self.group = group
        self.filename = filename
        self.extension = extension


def get_command_map(command_dirs):
    """
    Return a dictionary mapping command names to their Python module directory.
    The dictionary is in the format {command_name: module_name}.
    """
    from dodo_commands.framework.python_command_handler import PythonCommandHandler
    from dodo_commands.framework.yaml_command_handler import YamlCommandHandler
    from dodo_commands.framework.shell_command_handler import ShellCommandHandler

    command_map = {}

    file_map = {}
    for command_dir in command_dirs:
        file_map[command_dir] = list(
            glob.glob(os.path.join(command_dir, '*.*')))

    for handler in (PythonCommandHandler(), YamlCommandHandler(),
                    ShellCommandHandler()):
        handler.add_commands_to_map(command_dirs, file_map, command_map)

    return command_map


def execute_script(command_map, command_name):
    """
    Executes the script associated with command_name by importing its package.
    The script is assumed to have an entry point that is executed if
    Dodo.is_main(__name__) is True.
    """
    from dodo_commands.framework.python_command_handler import PythonCommandHandler
    from dodo_commands.framework.yaml_command_handler import YamlCommandHandler
    from dodo_commands.framework.shell_command_handler import ShellCommandHandler

    command_map_item = command_map[command_name]

    if command_map_item.extension == 'py':
        PythonCommandHandler().execute(command_map_item, command_name)
    elif command_map_item.extension == 'yaml':
        YamlCommandHandler().execute(command_map_item, command_name)
    elif command_map_item.extension == 'sh':
        ShellCommandHandler().execute(command_map_item, command_name)
    else:
        raise Exception("Logical error")



import glob
import os
import sys
from argparse import ArgumentParser

from dodo_commands import Dodo
from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.container.utils import get_ordered_layer_paths
from dodo_commands.framework.decorator_utils import _all_decorators
from dodo_commands.framework.paths import Paths


def _args():  # noqa
    parser = ArgumentParser(
        description=('Print the value of /ROOT/<what>_dir. For example: ' +
                     '"dodo which src" prints the value of /ROOT/src_dir.'))
    parser.add_argument(
        'what',
        nargs='?',
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--config',
                       action="store_true",
                       help='Print where the config file is')
    group.add_argument('--global-config',
                       action="store_true",
                       help='Print where the global config file is')
    group.add_argument('--env-dir',
                       action="store_true",
                       help='Print where the environment directory is')
    group.add_argument('--project-dir',
                       action="store_true",
                       help='Print where the project directory is')
    group.add_argument('--config-dir',
                       action="store_true",
                       help='Print where the config directory is')
    group.add_argument('--python-env-dir',
                       action="store_true",
                       help='Print where the python virtual env directory is')
    group.add_argument(
        '--script',
        help='Print where the dodo command script with given name is')
    group.add_argument('--decorators',
                       action="store_true",
                       help='Prints which command decorators are available')
    group.add_argument('--envs',
                       action="store_true",
                       help='Prints which environments are available')
    group.add_argument('--env',
                       action="store_true",
                       help='Prints name of the active environment')
    group.add_argument('--layers',
                       action="store_true",
                       help='Prints which layers are active')
    group.add_argument(
        '--default-commands',
        action="store_true",
        help='Prints the directory where default commands are stored')
    group.add_argument(
        '--global-commands',
        action="store_true",
        help='Prints the directory where global commands are stored')

    return Dodo.parse_args(parser)


def _which_script(script):
    command_dirs = get_command_dirs_from_config(Dodo.get_config())
    for item in command_dirs:
        script_path = os.path.join(item, script + ".py")
        if os.path.exists(script_path):
            return os.path.realpath(script_path)

    for item in command_dirs:
        for yml in glob.glob(os.path.join(item, "dodo.*.yaml")):
            with open(yml) as ifs:
                if script in yaml_round_trip_load(ifs.read()).keys():
                    return os.path.realpath(yml)

    return None


def _which_dir(directory):
    return Dodo.get_config("/ROOT/%s_dir" % directory, None)


if Dodo.is_main(__name__):
    args = _args()

    if args.config:
        if os.path.exists(Paths().config_dir()):
            print(os.path.join(Paths().config_dir(), "config.yaml"))
    elif args.global_config:
        sys.stdout.write(Paths().global_config_filename() + '\n')
    elif args.global_commands:
        sys.stdout.write(Paths().global_commands_dir() + '\n')
    elif args.default_commands:
        sys.stdout.write(Paths().default_commands_dir() + '\n')
    elif args.script:
        sys.stdout.write(_which_script(args.script) + '\n')
    elif args.env_dir:
        sys.stdout.write(Paths().env_dir())
    elif args.project_dir:
        sys.stdout.write(_which_dir("project") + '\n')
    elif args.config_dir:
        sys.stdout.write(_which_dir("config") + '\n')
    elif args.python_env_dir:
        python_env_dir = os.path.realpath(
            os.path.join(Paths().env_dir(), "python_env_dir"))
        if os.path.exists(python_env_dir):
            sys.stdout.write(python_env_dir + '\n')
    elif args.decorators:
        sys.stdout.write(
            ", ".join(sorted(_all_decorators(Dodo.get_config()).keys())) +
            '\n')
    elif args.envs:
        sys.stdout.write('\n'.join(sorted(os.listdir(Paths().envs_dir()))) +
                         '\n')
    elif args.env:
        if Dodo.get_config("/ROOT/env_name", None):
            sys.stdout.write(Dodo.get_config("/ROOT/env_name") + '\n')
    elif args.layers:
        layer_paths = get_ordered_layer_paths(Dodo.get_container())
        for layer_path in layer_paths:
            sys.stdout.write(layer_path + '\n')
    elif args.what:
        x = _which_script(args.what)
        if x:
            sys.stdout.write(x + '\n')
    else:
        if Dodo.get_config("/ROOT/env_name", None):
            sys.stdout.write(Dodo.get_config("/ROOT/env_name") + '\n')



import argparse
import os
import sys
from argparse import ArgumentParser

from dodo_commands import Dodo
from dodo_commands.framework.command_map import get_command_map, execute_script
from dodo_commands.framework.command_path import get_command_dirs


def _args():
    parser = ArgumentParser(description='Download a command script and run it')

    parser.add_argument('url')
    parser.add_argument('script_args', nargs=argparse.REMAINDER)

    args = Dodo.parse_args(parser)
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    tmp_dir = os.path.join(os.path.expanduser('~'), '.dodo_commands', 'tmp')
    package_dir = os.path.join(tmp_dir, 'exec_scripts')

    if '.commands.yaml' in args.url:
        last_slash = args.url.rfind('/')
        url = args.url[:last_slash]
        cmd_name = args.url[last_slash + 1:]
    else:
        url = args.url
        cmd_name = os.path.splitext(os.path.basename(url))[0]

    Dodo.run(['mkdir', '-p', package_dir])
    Dodo.run(['touch', os.path.join(package_dir, '__init__.py')])
    Dodo.run(
        ['wget', url, '-O',
         os.path.join(package_dir, os.path.basename(url))])

    command_map = get_command_map(get_command_dirs([package_dir], []))
    sys.argv = sys.argv[1:]
    execute_script(command_map, cmd_name)



"""Tests for config.py."""

from ..config import ConfigExpander


class TestConfigExpander:  # noqa
    simple_config = {
        'ROOT': {
            'command_path': ['foo', 'bar']
        },
        'FOO': {
            'bar': {
                'one': 'two'
            },
            'one': [1, 2, 3],
            'two': 2.2
        }
    }

    config_with_refs = {
        'ROOT': {
            'command_path': ['foo $PATH', 'bar']
        },
        'FOO': {
            'bar': {
                'one': '${/ROOT/command_path/1} and ${/FOO/two}'
            },
            'one': [1, 2, 3],
            'two': '${/FOO/one/2}',
            'three': '${/FOO/one}'
        }
    }

    def test_config_expander(self):
        """Config values are converted into environment variables."""
        config_expander = ConfigExpander()
        config_expander.run(self.simple_config)

    def test_expand(self):
        """dodo/system environment variables are expanded."""
        config_expander = ConfigExpander()
        config_expander.run(self.config_with_refs)
        assert self.config_with_refs['FOO']['bar']['one'] == 'bar and 3'
        assert self.config_with_refs['FOO']['three'] == '[1, 2, 3]'
        assert "$PATH" not in self.config_with_refs['ROOT']['command_path'][0]



from dodo_commands.dependencies.get import funcy
from dodo_commands.framework.global_config import load_global_config_parser

merge = funcy.merge


class Commands:
    def __init__(self):
        self.layer_name_by_inferred_command = {}
        self.global_aliases = {}
        self.aliases_from_config = {}
        self.command_map = {}
        self.command_dirs = []

    @property
    def aliases(self):
        return merge(self.global_aliases, self.aliases_from_config)

    @staticmethod
    def get(ctr):
        return ctr.commands


def init_commands(self):
    global_config = load_global_config_parser()
    if global_config.has_section('alias'):
        self.global_aliases = dict(global_config.items('alias'))
    return self



from dodo_commands.framework.container import actions, facets
from dodo_commands.framework.paths import Paths


class Container:
    def _create_facets(self):
        self.paths = Paths()
        self.command_line = facets.init_command_line(facets.CommandLine())
        self.layers = facets.init_layers(facets.Layers(), 'config.yaml')
        self.config = facets.Config()
        self.commands = facets.init_commands(facets.Commands())

    def __init__(self):
        self._create_facets()

    def run_actions(self):
        actions.layers.action_load_root_layer(self)
        actions.layers.action_load_named_layers(self)

        actions.commands.action_get_inferred_command_map(self)

        actions.command_line.action_get_expanded_layer_paths(self)
        actions.command_line.action_get_inferred_layer_paths(self)

        # Create the config
        actions.layers.action_select_layers(self)
        actions.config.action_check_conflicts_in_selected_layer_paths(self)
        actions.config.action_build_from_selected_layers(self)

        actions.commands.action_get_aliases_from_config(self)
        actions.commands.action_get_command_map(self)

        # This step may find out that the selected command alias
        # contains some --layer=foo.bar.yaml terms. These additional
        # layers are stored in self.command_line.more_given_layer_paths
        actions.command_line.action_expand_and_autocomplete_command_name(self)

        if self.command_line.more_given_layer_paths:
            actions.layers.action_select_layers(self)
            actions.config.action_build_from_selected_layers(self)



from dodo_commands.dependencies.get import funcy

filter, map, partial, rcompose, remove = (
    funcy.filter,
    funcy.map,
    funcy.partial,
    funcy.rcompose,
    funcy.remove,
)


def keep_if(f):
    return partial(filter, f)


def keep_truthy():
    return keep_if(bool)


def str_split_at(s, pos):
    return s[:pos], s[pos:]


def remove_if(f):
    return partial(remove, f)


def map_with(f):
    return partial(map, f)


def for_each(op):
    def f(seq):
        for x in seq:
            op(x)
        return seq

    return f


def pipe(x, *functions):
    return rcompose(*functions)(x)


def negate(f):
    return lambda x: not f(x)


def ds(f):
    return lambda x: f(*x)


def debug(x):
    print(x)
    __import__('pud' + 'b').set_trace()
    return x


def debug_as(f):
    return lambda x: debug(f(x))


def drill(x, *keys, default=None):
    result = x
    for key in keys:
        if not result or key not in result:
            return default
        result = result[key]
    return result


def map_to_tuple(f):
    def result(*args, **kwargs):
        return (f(*args, **kwargs), )

    return result



import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import funcy, six
from dodo_commands.framework.config_expander import Key
from dodo_commands.framework.config_io import ConfigIO
from dodo_commands.framework.container.utils import get_ordered_layer_paths
from dodo_commands.framework.global_config import load_global_config_parser
from dodo_commands.framework.paths import Paths

concat = funcy.concat
configparser = six.moves.configparser


def _args():
    parser = ArgumentParser(description=('Edit the dodo configuration files'))

    parser.add_argument(
        '--key', help='Only edit this key. Used in conjunction with --val')
    parser.add_argument(
        '--val',
        help='The value to be used in conjunction with the --key option')

    args = Dodo.parse_args(parser)

    if args.key or args.val:
        if not args.key and args.val:
            raise CommandError(
                "The options --key and --val should always be used together")

    args.editor = load_global_config_parser().get("settings", "config_editor")
    args.config_dir = Dodo.get_config('/ROOT/config_dir')
    return args


if Dodo.is_main(__name__, safe=('--key' not in sys.argv)):
    try:
        args = _args()
    except configparser.NoOptionError as e:
        raise CommandError("{error}. Please check {filename}".format(
            error=str(e), filename=Paths().global_config_filename()))

    config = ConfigIO().load()

    if args.key and args.val:
        key = Key(config, args.key)
        key.set(args.val)
        ConfigIO().save(config)
        sys.exit(0)

    def add_global_config_filename(layer_paths):
        return concat(layer_paths, [Paths().global_config_filename()])

    x = get_ordered_layer_paths(Dodo.get_container())
    x = add_global_config_filename(x)
    yaml_filenames = x

    Dodo.run(args.editor.split() + yaml_filenames, cwd='.')



import os
import platform
import re
import subprocess
from importlib import import_module

from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.command_map import CommandMapItem
from dodo_commands.framework.command_path import extend_sys_path
from dodo_commands.framework.config import Paths
from dodo_commands.framework.util import query_yes_no

CommandNotFound = plumbum.commands.processes.CommandNotFound


class PythonCommandMapItem(CommandMapItem):
    def __init__(self, group, filename):
        super(PythonCommandMapItem, self).__init__(group, filename, 'py')
        self.package_path = group


class PythonCommandHandler:
    def add_commands_to_map(self, command_dirs, file_map, command_map):
        extend_sys_path(command_dirs)
        for command_dir, files in file_map.items():
            for file in files:
                command_name, ext = os.path.splitext(os.path.basename(file))
                if ext == '.py' and not command_name.startswith('_'):
                    command_map[command_name] = PythonCommandMapItem(
                        group=os.path.basename(command_dir), filename=file)

    def execute(self, command_map_item, command_name):
        def meta_data_filename():
            base_path = import_module(
                command_map_item.package_path).__path__[0]
            return os.path.join(base_path, command_name + ".meta")

        def meta_data():
            if os.path.exists(meta_data_filename()):
                with open(meta_data_filename()) as f:
                    return yaml_round_trip_load(f.read())
            return None

        def install_packages(meta_data, dependency):
            """Pip install packages found in meta_data."""
            packager_name = 'brew' if platform.system() == 'Darwin' else 'apt'
            package_data = meta_data['packages'].get(packager_name, {})
            recipe = package_data.get(dependency, None)
            if recipe:
                commands = [recipe] if isinstance(recipe, str) else recipe

                print("This command wants to run:\n%s\n" % '\n'.join(commands))
                if query_yes_no("Install (yes), or abort (no)?"):
                    for package_cmd in commands:
                        subprocess.check_call(package_cmd.split())
                    print("--- Done ---\n\n")

        def install_python_packages(meta_data, dependency):
            """Pip install packages found in meta_data."""
            requirements = [
                x for x in meta_data['requirements']
                if re.search(r"\b" + dependency + r"\b", x)
            ]
            if requirements:
                if len(requirements) > 1:
                    raise CommandError(
                        "Found more than 1 candidate for python package %s in %s"
                        % (dependency, meta_data_filename()))

                requirement = requirements[0]
                print("This command wants to install %s:\n" % requirement)
                if query_yes_no("Install (yes), or abort (no)?"):
                    subprocess.check_call(
                        [sys.executable, "-m", "pip", "install", requirement])
                    print("--- Done ---\n\n")

        def get_program_from_exception(e):
            m = re.search(r"cannot import name '(\w+)' from 'plumbum.cmd'",
                          e.args[0]) if len(e.args) else None
            return m.groups(1)[0] if m else None

        import_path = '%s.%s' % (command_map_item.package_path, command_name)
        if command_map_item.package_path in ("", None, "."):
            import_path = command_name

        # Try the import, return if success
        try:
            return import_module(import_path)
        except ImportError as e:
            exception_type = type(e)
            program = get_program_from_exception(e)
            if program:
                dependency, install = program, install_packages
            else:
                dependency, install = e.name, install_python_packages
        except CommandNotFound as e:
            exception_type = type(e)
            dependency, install = e.name, install_python_packages

        # Try the import again, but first run install
        try:
            md = meta_data()
            if md:
                install(md, dependency)
            import_module(import_path)
        except exception_type:
            fail_msg = ('Could not find executable: %s' % dependency if
                        (install is install_packages) else
                        'Could not import python module: %s' % dependency)
            raise CommandError(fail_msg)



from .layers import Layers, init_layers  # noqa
from .command_line import CommandLine, init_command_line  # noqa
from .config import Config  # noqa
from .commands import Commands, init_commands  # noqa
from dodo_commands.framework.funcy import (ds, for_each, keep_if)


def i_(facet_class, member, prefix=None, alt_name=None):
    arg_name = (alt_name if alt_name else prefix + '_' +
                member if prefix else member)
    return ('in', facet_class, member, arg_name)


def o_(facet_class, member):
    return ('out', facet_class, member)


def map_datas(*args, transform=None):
    def action(ctr):
        kwargs = {}

        def is_input(arg):
            return arg[0] == 'in'

        def do_add_to_kwargs(input_arg):
            _, facet_class, member, arg_name = input_arg
            value = getattr(facet_class.get(ctr), member)
            kwargs[arg_name] = value

        x = args
        # [(in | out, facet_class, member)]
        x = keep_if(is_input)(x)
        # [(in | out, facet_class, member)]
        x = for_each(do_add_to_kwargs)(x)
        # [value]
        output_values = transform(**kwargs)

        def is_output(arg):
            return arg[0] == 'out'

        def zip_with_output_values(output_args):
            assert len(output_args) == len(output_values)
            return zip(output_args, output_values)

        def do_store(output_arg, output_value):
            _, facet_class, member = output_arg
            setattr(facet_class.get(ctr), member, output_value)

        x = args
        # [(in | out, facet_class, member)]
        x = keep_if(is_output)(x)
        # [(in | out, facet_class, member)]
        x = zip_with_output_values(x)
        # [((in | out, facet_class, member), output_value)]
        for_each(ds(do_store))(x)

    return action



import os
from argparse import ArgumentParser

from dodo_commands import Dodo


def _args():
    parser = ArgumentParser()
    parser.add_argument('container_type',
                        choices=Dodo.get_config('/DOCKER/containers').keys())
    parser.add_argument('output_dir')
    parser.add_argument('--reverse', action='store_true')
    args = Dodo.parse_args(parser)
    return args


def _docker_run(output_dir, args):
    Dodo.run([
        'docker', 'run', '--rm', '--volumes-from', container_name, '--volume',
        '%s:/tmp/docker-snapshot' % output_dir, container_type['image']
    ] + args,
             cwd='.')


if Dodo.is_main(__name__, safe=True):
    args = _args()
    container_name = Dodo.get_config('/DOCKER/containers/' +
                                     args.container_type)

    docker_options = Dodo.get_config('/DOCKER_OPTIONS', {})
    docker_options.setdefault('docker-snapshot', {})
    docker_options['docker-snapshot']['volumes_from'] = container_name

    container_type = Dodo.get_config('/DOCKER/container_types/' +
                                     args.container_type)
    for path in container_type['dirs']:
        host_output_path = os.path.join(args.output_dir, path[1:])
        docker_output_path = os.path.join('/tmp/docker-snapshot', path[1:])

        if args.reverse:
            _docker_run(args.output_dir,
                        ['rm', '-rf', os.path.join(path, '*')])
            _docker_run(
                args.output_dir,
                ['cp', '-rf',
                 os.path.join(docker_output_path, '*'), path])
        else:
            if os.path.exists(host_output_path):
                Dodo.run(['rm', '-rf', host_output_path], cwd='.')
            Dodo.run(['mkdir', '-p',
                      os.path.dirname(host_output_path)],
                     cwd='.')
            _docker_run(
                args.output_dir,
                ['cp', '-rf', path,
                 os.path.dirname(docker_output_path)])



from dodo_commands.framework.command_error import CommandError  # noqa
from dodo_commands.framework.config_io import ConfigIO
from dodo_commands.framework.funcy import drill


class LayerConfig:
    def __init__(self, target_path, inferred_commands, group_name):
        self.target_path = target_path
        self.inferred_commands = inferred_commands
        self.group_name = group_name


class Layers:
    def __init__(self):
        self.config_io = ConfigIO()
        self.root_layer_path = None
        self.root_layer = None
        self.layer_by_target_path = {}
        self.selected_layer_by_path = {}

    @property
    def layer_props_by_layer_name(self):
        result = {}

        groups = drill(self.root_layer, 'LAYER_GROUPS', default={})
        for group_name, group in groups.items():
            for group_item in group:

                if isinstance(group_item, str):
                    layer_name = group_item
                    base_name = layer_name
                    inferred_commands = []
                    target_path = None
                elif isinstance(group_item, dict):
                    layer_name, layer_props = list(group_item.items())[0]
                    layer_props = layer_props or {}
                    inferred_commands = list(layer_props.get(
                        'inferred_by', []))
                    target_path = layer_props.get('target_path')
                    base_name = layer_props.get('base_name', layer_name)

                target_path = target_path or "%s.%s.yaml" % (group_name,
                                                             base_name)

                layer_props = LayerConfig(target_path, inferred_commands,
                                          group_name)

                if layer_name in result:
                    prev_layer_props = result[layer_name]
                    raise CommandError(
                        "Name conlict for layer '%s' in groups '%s' and '%s'" %
                        (layer_name, prev_layer_props.group_name,
                         layer_props.group_name))

                result[layer_name] = layer_props
        return result

    @staticmethod
    def get(ctr):
        return ctr.layers


def init_layers(self, root_layer_path):
    self.root_layer_path = root_layer_path
    return self



import os
import subprocess
import sys
from distutils.command.install_data import install_data

from setuptools import find_packages, setup


class InstallPrivatePackages(install_data):
    def _packages_dirname(self):
        def find_module_path():
            for p in sys.path:
                if os.path.isdir(p) and "dodo_commands" in os.listdir(p):
                    return os.path.join(p, "dodo_commands")

        install_path = find_module_path()
        return os.path.join(install_path, "dependencies", "packages")

    def _install_packages(self, packages_dirname):
        if not os.path.exists(packages_dirname):
            os.makedirs(packages_dirname)

        for dependency in [
                'python-dotenv==0.12.0',
                'plumbum==1.6.8',
                'ruamel.yaml==0.16.10',
                'parsimonious==0.8.1',
                'six==1.14.0',
                'funcy==1.14',
                'ansimarkup==1.4.0',
                'argcomplete==1.11.1',
                'semantic_version==2.8.4',
        ]:
            try:
                subprocess.check_call([
                    sys.executable, '-m', 'pip', 'install', '--target',
                    packages_dirname, dependency
                ])
            except:  # noqa
                pass

    def run(self):
        super().run()
        self._install_packages(self._packages_dirname())


setup(name='dodo_commands',
      version='0.30.4',
      description=
      'Project-aware development environments, inspired by django-manage',
      url='https://github.com/mnieber/dodo_commands',
      download_url='https://github.com/mnieber/dodo_commands/tarball/0.30.4',
      author='Maarten Nieber',
      author_email='hallomaarten@yahoo.com',
      license='MIT',
      packages=find_packages(),
      package_data={
          'dodo_commands': [
              'extra/dodo_standard_commands/*.meta',
              'extra/dodo_docker_commands/*.meta',
          ]
      },
      entry_points={'console_scripts': [
          'dodo=dodo_commands.dodo:main',
      ]},
      data_files=[('/etc/bash_completion.d', [
          'dodo_commands/bin/dodo_autocomplete.sh',
          'dodo_commands/bin/sdodo_autocomplete.sh'
      ]), ('/etc/fish/conf.d', ['dodo_commands/bin/dodo_autocomplete.fish'])],
      cmdclass={'install_data': InstallPrivatePackages},
      install_requires=[],
      zip_safe=False)



from collections import OrderedDict

from dodo_commands.dependencies.get import funcy
from dodo_commands.framework.alias import load_named_layers
from dodo_commands.framework.config_layers import layer_filename_superset
from dodo_commands.framework.container.facets import (CommandLine, Layers, i_,
                                                      map_datas, o_)

distinct = funcy.distinct


# LAYERS
def action_load_root_layer(ctr):
    def transform(
        config_io,
        root_layer_path,
    ):
        if ctr.paths.project_dir():
            return (config_io.load(root_layer_path), )
        return ({}, )

    return map_datas(i_(Layers, 'config_io'),
                     i_(Layers, 'root_layer_path'),
                     o_(Layers, 'root_layer'),
                     transform=transform)(ctr)


# LAYERS
def action_load_named_layers(ctr):
    def transform(
        config_io,
        layer_props_by_layer_name,
    ):
        layer_by_target_path = load_named_layers(
            config_io,
            layer_props_by_layer_name,
        )
        return (layer_by_target_path, )

    return map_datas(i_(Layers, 'config_io'),
                     i_(Layers, 'layer_props_by_layer_name'),
                     o_(Layers, 'layer_by_target_path'),
                     transform=load_named_layers)(ctr)


# LAYERS
def action_select_layers(ctr):
    def transform(
        config_io,
        root_layer_path,
        root_layer,
        command_line_layer_paths,
    ):
        all_layer_paths = layer_filename_superset(
            distinct([root_layer_path] + command_line_layer_paths),
            config_io=config_io)

        selected_layer_by_path = OrderedDict()
        for layer_path in all_layer_paths:
            selected_layer_by_path[layer_path] = config_io.load(layer_path)
        return (selected_layer_by_path, )

    return map_datas(i_(Layers, 'config_io'),
                     i_(Layers, 'root_layer_path'),
                     i_(Layers, 'root_layer'),
                     i_(CommandLine,
                        'layer_paths',
                        alt_name='command_line_layer_paths'),
                     o_(Layers, 'selected_layer_by_path'),
                     transform=transform)(ctr)



from dodo_commands.dependencies.get import funcy
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.funcy import drill, for_each

distinct, merge = funcy.distinct, funcy.merge


def get_inferred_command_map(
    global_aliases,
    layer_props_by_layer_name,
    layer_by_target_path,
):
    layer_name_by_inferred_command = {}

    for layer_name, layer_props in layer_props_by_layer_name.items():

        def get_layer(layer_props):
            return layer_by_target_path[layer_props.target_path]

        def get_cmd_aliases(layer):
            command_aliases = drill(layer, 'ROOT', 'aliases', default={})
            for alias, target in global_aliases.items():
                if alias in command_aliases:
                    raise CommandError(
                        "Ambigous command alias %s, could be %s or %s" %
                        (alias, target, command_aliases[alias]))
            return merge(command_aliases, global_aliases)

        def get_inferred_aliases(command_aliases, inferred_commands):
            return list(alias for alias in command_aliases.keys()
                        if command_aliases[alias] in inferred_commands)

        def get_inferred_commands_and_aliases(inferred_commands,
                                              inferred_aliases):
            return distinct(inferred_commands + inferred_aliases)

        def do_store(inferred_command):
            if inferred_command in layer_name_by_inferred_command:
                raise CommandError(
                    "Ambigous inferred command %s. " % inferred_command +
                    "Could be %s or %s" %
                    (layer_props.target_path,
                     layer_name_by_inferred_command[inferred_command]))
            layer_name_by_inferred_command[inferred_command] = layer_name

        x = layer_props
        x = layer = get_layer(x)
        x = command_aliases = get_cmd_aliases(layer)
        x = inferred_aliases = get_inferred_aliases(
            command_aliases, layer_props.inferred_commands)
        x = get_inferred_commands_and_aliases(layer_props.inferred_commands,
                                              inferred_aliases)
        for_each(do_store)(x)

    return layer_name_by_inferred_command
